from solaredge2mqtt.models.base import Component, EnumModel
from solaredge2mqtt.models.energy import Energy, EnergyPeriod, EnergyQuery
from solaredge2mqtt.models.modbus import (SunSpecBattery, SunSpecInverter,
                                          SunSpecMeter)
from solaredge2mqtt.models.monitoring import (LogicalInfo, LogicalInverter,
                                              LogicalModule, LogicalString)
from solaredge2mqtt.models.powerflow import Powerflow
from solaredge2mqtt.models.wallbox import WallboxAPI



from typing import Dict
from solaredge2mqtt.models.base import Component


class WallboxAPI(Component):
    COMPONENT = "wallbox"
    SOURCE = "api"

    power: float
    state: str
    vehicle_plugged: bool
    max_current: float

    def __init__(self, data: Dict[str, str | int]):
        power = round(data["meter"]["totalActivePower"] / 1000)
        state = data["state"]
        vehicle_connected = bool(data["vehiclePlugged"])
        max_current = float(data["maxCurrent"])

        super().__init__(
            power=power,
            state=state,
            vehicle_plugged=vehicle_connected,
            max_current=max_current,
        )



import json
from typing import Optional

import requests
from requests.exceptions import HTTPError, RequestException

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import (
    LogicalInfo,
    LogicalInverter,
    LogicalModule,
    LogicalString,
)
from solaredge2mqtt.settings import ServiceSettings

LOGIN_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/login"
LOGICAL_URL = "https://monitoring.solaredge.com/solaredge-apigw/api/sites/{site_id}/layout/logical"


class MonitoringSite:
    def __init__(self, settings: ServiceSettings) -> None:
        self.username = settings.api_username
        self.password = settings.api_password
        self.site_id = settings.api_site_id

        self.session = requests.session()

    def login(self) -> None:
        result = self.session.post(
            LOGIN_URL,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={"j_username": self.username, "j_password": self.password},
        )

        if result.status_code == 200:
            logger.info("Login to monitoring site successful")
        else:
            raise HTTPError(
                f"Cannot login to monitoring site (status code: {result.status_code})"
            )

    def get_module_energies(self) -> list[LogicalInverter] | None:
        modules = None
        try:
            logical = self._get_logical()

            logger.debug("Logical: {logical}", logical=json.dumps(logical, indent=4))

            inverters = self._parse_inverters(
                logical["logicalTree"]["children"], logical["reportersData"]
            )

            modules = []

            for inverter in inverters:
                logger.debug(
                    "Inverter: {inverter}", inverter=inverter.model_dump_json(indent=4)
                )
                for string in inverter.strings:
                    for module in string.modules:
                        modules.append(module)
        except RequestException as error:
            logger.error(
                "Exception while reading data from monitoring site: {exception}",
                exception=error,
            )

        return modules

    def _get_logical(self, run: Optional[int] = 0) -> dict:
        result = self.session.get(
            LOGICAL_URL.format(site_id=self.site_id),
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "X-CSRF-TOKEN": self.session.cookies["CSRF-TOKEN"],
            },
        )

        if result.status_code == 200:
            return json.loads(result.content.decode("utf-8"))
        elif result.status_code in (401, 403) and run == 0:
            logger.info("Login expired, logging in again")
            self.login()
            logger.info("Retrying to get logical layout")
            return self._get_logical(run=1)

        raise HTTPError(
            f"Cannot read logical layout (status code: {result.status_code})"
        )

    def _parse_inverters(self, inverter_objs, reporters_data) -> list[LogicalInverter]:
        inverters = []

        for inverter_obj in inverter_objs:
            info = LogicalInfo.map(inverter_obj["data"])
            if "INVERTER" in info["type"]:
                inverter = LogicalInverter(
                    info=info,
                    energy=reporters_data[info["id"]]["unscaledEnergy"]
                    if info["id"] in reporters_data
                    else None,
                )

                self._parse_strings(inverter, inverter_obj["children"], reporters_data)

                inverters.append(inverter)

            else:
                logger.info("Unknown inverter type: {type}", type=info["type"])

        return inverters

    def _parse_strings(self, inverter, string_objs, reporters_data):
        for string_obj in string_objs:
            info = LogicalInfo.map(string_obj["data"])
            string = LogicalString(
                info=info,
                energy=reporters_data[info["id"]]["unscaledEnergy"]
                if info["id"] in reporters_data
                else None,
            )

            self._parse_panels(string, string_obj["children"], reporters_data)

            inverter.strings.append(string)

    def _parse_panels(self, string, panel_objs, reporters_data):
        for panel_obj in panel_objs:
            info = LogicalInfo.map(panel_obj["data"])
            panel = LogicalModule(
                info=info,
                energy=reporters_data[info["id"]]["unscaledEnergy"]
                if info["id"] in reporters_data
                else None,
            )

            string.modules.append(panel)



from __future__ import annotations

from typing import Dict, Optional

from pydantic import BaseModel


class LogicalInfo(BaseModel):
    id: str
    serialnumber: Optional[str]
    name: str
    type: str

    @staticmethod
    def map(data: Dict[str, str | int]) -> Dict[str, str]:
        return {
            "id": str(data["id"]),
            "serialnumber": data["serialNumber"],
            "name": data["name"],
            "type": data["type"],
        }


class LogicalInverter(BaseModel):
    info: LogicalInfo
    energy: Optional[float]
    strings: list[LogicalString] = []


class LogicalString(BaseModel):
    info: LogicalInfo
    energy: Optional[float]
    modules: list[LogicalModule] = []


class LogicalModule(BaseModel):
    info: LogicalInfo
    energy: Optional[float]



from __future__ import annotations

from datetime import datetime

from pydantic import computed_field

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models.base import EnumModel, Solaredge2MQTTBaseModel


class EnergyQuery(EnumModel):
    ACTUAL = "energy_actual_unit"
    LAST = "energy_last_unit"

    def __init__(self, query: str) -> None:
        self._query: str = query

    @property
    def query(self) -> str:
        return self._query


class EnergyPeriod(EnumModel):
    TODAY = "today", "1d", EnergyQuery.ACTUAL
    YESTERDAY = "yesterday", "1d", EnergyQuery.LAST
    THIS_WEEK = "this_week", "1w", EnergyQuery.ACTUAL
    LAST_WEEK = "last_week", "1w", EnergyQuery.LAST
    THIS_MONTH = "this_month", "1mo", EnergyQuery.ACTUAL
    LAST_MONTH = "last_month", "1mo", EnergyQuery.LAST
    THIS_YEAR = "this_year", "1y", EnergyQuery.ACTUAL

    def __init__(self, topic: str, unit: str, query: EnergyQuery) -> None:
        self._topic: str = topic
        self._unit: str = unit
        self._query: EnergyQuery = query

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def unit(self) -> str:
        return self._unit

    @property
    def query(self) -> EnergyQuery:
        return self._query


class Energy(Solaredge2MQTTBaseModel):
    info: EnergyInfo
    pv_production: float
    inverter: InverterEnergy
    grid: GridEnergy
    battery: BatteryEnergy
    consumer: ConsumerEnergy

    def __init__(
        self,
        energy_data: dict,
        period: EnergyPeriod,
    ):
        logger.trace(energy_data)
        pv_production = energy_data["pv_production"]

        subclass_values: dict[str, dict[str, float]] = {}

        for key, value in energy_data.items():
            keys = key.split("_")
            if len(keys) < 2:
                continue

            if keys[0] in ["consumer", "inverter", "grid", "battery"]:
                if keys[0] not in subclass_values:
                    subclass_values[keys[0]] = {}

                subclass_values[keys[0]]["_".join(keys[1:])] = round(value, 3)

        logger.debug(subclass_values["consumer"])

        super().__init__(
            pv_production=round(pv_production, 3),
            inverter=InverterEnergy(**subclass_values["inverter"]),
            grid=GridEnergy(**subclass_values["grid"]),
            battery=BatteryEnergy(**subclass_values["battery"]),
            consumer=ConsumerEnergy(**subclass_values["consumer"]),
            info=EnergyInfo(
                period=period, start=energy_data["_start"], stop=energy_data["_stop"]
            ),
        )

    @computed_field
    @property
    def self_consumption_rates(self) -> SelfConsumptionRate:
        return SelfConsumptionRate(self)

    @computed_field
    @property
    def self_sufficiency_rates(self) -> SelfSufficiencyRate:
        return SelfSufficiencyRate(self)


class EnergyInfo(Solaredge2MQTTBaseModel):
    period: EnergyPeriod
    start: datetime
    stop: datetime


class InverterEnergy(Solaredge2MQTTBaseModel):
    production: float
    consumption: float
    dc_power: float
    pv_production: float
    battery_production: float


class GridEnergy(Solaredge2MQTTBaseModel):
    delivery: float
    consumption: float


class BatteryEnergy(Solaredge2MQTTBaseModel):
    charge: float
    discharge: float


class ConsumerEnergy(Solaredge2MQTTBaseModel):
    house: float
    evcharger: float
    inverter: float

    total: float

    used_production: float
    used_pv_production: float
    used_battery_production: float


class SelfConsumptionRate(Solaredge2MQTTBaseModel):
    grid: int
    battery: int
    pv: int
    total: int

    def __init__(self, energy: Energy):
        if energy.inverter.production > 0:
            grid_rate = int(
                round(energy.grid.delivery / energy.inverter.production * 100)
            )
            battery_rate = int(
                round(
                    energy.consumer.used_battery_production
                    / energy.inverter.production
                    * 100
                )
            )
            pv_rate = 100 - grid_rate - battery_rate
            total = battery_rate + pv_rate
        else:
            grid_rate = 0
            battery_rate = 0
            pv_rate = 0
            total = 0

        super().__init__(
            grid=grid_rate,
            battery=battery_rate,
            pv=pv_rate,
            total=total,
        )


class SelfSufficiencyRate(Solaredge2MQTTBaseModel):
    grid: int
    battery: int
    pv: int
    total: int

    def __init__(self, energy: Energy):
        if energy.consumer.total > 0:
            grid_rate = int(
                round(energy.grid.consumption / energy.consumer.total * 100)
            )
            battery_rate = int(
                round(
                    energy.consumer.used_battery_production
                    / energy.consumer.total
                    * 100
                )
            )
            pv_rate = 100 - grid_rate - battery_rate
            total = battery_rate + pv_rate
        else:
            grid_rate = 0
            battery_rate = 0
            pv_rate = 0
            total = 0

        super().__init__(
            grid=grid_rate,
            battery=battery_rate,
            pv=pv_rate,
            total=total,
        )



"""
    This module, service.py, is part of the SolarEdge2MQTT service, which reads data 
    from a SolarEdge inverter and publishes it to an MQTT broker. It uses the asyncio 
    library for asynchronous I/O and the aiomqtt library for MQTT communication. 
    The module also includes a run function to initialize and start the service.
"""
import asyncio as aio
import signal

from aiomqtt import MqttError

from solaredge2mqtt.api import MonitoringSite
from solaredge2mqtt.logging import initialize_logging, logger
from solaredge2mqtt.modbus import Modbus
from solaredge2mqtt.models import EnergyPeriod, EnergyQuery, Powerflow
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.persistence.influxdb import InfluxDB
from solaredge2mqtt.settings import service_settings
from solaredge2mqtt.wallbox import WallboxClient


def run():
    service = Service()
    loop = aio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, service.cancel)
    loop.add_signal_handler(signal.SIGTERM, service.cancel)
    loop.run_until_complete(service.main_loop())
    loop.close()


class Service:
    def __init__(self):
        self.settings = service_settings()
        self.mqtt = MQTTClient(self.settings)

        self.modbus = Modbus(self.settings)

        self.cancel_request = aio.Event()
        self.loops: set[aio.Task] = set()

        self.wallbox: WallboxClient | None = (
            WallboxClient(self.settings)
            if self.settings.is_wallbox_configured
            else None
        )

        self.influxdb: InfluxDB | None = (
            InfluxDB(self.settings) if self.settings.is_influxdb_configured else None
        )

        self.monitoring: MonitoringSite | None = (
            MonitoringSite(self.settings) if self.settings.is_api_configured else None
        )

    def cancel(self):
        logger.info("Stopping SolarEdge2MQTT service...")
        self.cancel_request.set()
        for loop in self.loops:
            loop.cancel()

    async def main_loop(self):
        initialize_logging(self.settings.logging_level)
        logger.info("Starting SolarEdge2MQTT service...")
        logger.debug(self.settings)

        logger.info("Timezone: {timezone}", timezone=self.settings.timezone)

        if self.settings.is_influxdb_configured:
            self.influxdb.initialize_buckets()
            self.influxdb.initialize_task()

        while not self.cancel_request.is_set():
            try:
                async with self.mqtt:
                    await self.mqtt.publish_status_online()

                    self.schedule_loop(self.settings.interval, self.basic_values_loop)

                    if self.settings.is_api_configured:
                        self.monitoring.login()
                        self.schedule_loop(300, self.monitoring_loop)

                    if self.settings.is_influxdb_configured:
                        self.schedule_loop(300, self.energy_loop)

                    await aio.gather(*self.loops)

                    await self.mqtt.publish_status_offline()

            except MqttError:
                logger.error("MQTT error, reconnecting in 5 seconds...")
                await aio.sleep(5)
            except aio.exceptions.CancelledError:
                logger.debug("Loops cancelled")
                return
            finally:
                await aio.sleep(2)
                for loop in self.loops:
                    loop.cancel()

    def schedule_loop(self, interval_in_seconds: int, handle: callable):
        loop = aio.create_task(self.run_loop(interval_in_seconds, handle))
        self.loops.add(loop)
        loop.add_done_callback(self.loops.remove)

    async def run_loop(self, interval_in_seconds: int, handle: callable):
        while not self.cancel_request.is_set():
            await handle()
            await aio.sleep(interval_in_seconds)

    async def basic_values_loop(self):
        results = await aio.gather(
            self.modbus.loop(),
            self.wallbox.loop()
            if self.settings.is_wallbox_configured
            else aio.sleep(0),
        )

        inverter_data, meters_data, batteries_data = results[0]

        if any(data is None for data in [inverter_data, meters_data, batteries_data]):
            logger.warning("Invalid modbus data, skipping this loop")
            return

        for battery in batteries_data.values():
            if not battery.is_valid:
                logger.warning("Invalid battery data, skipping this loop")
                logger.debug(battery)
                return

        wallbox_data = results[1]
        evcharger = 0

        if wallbox_data is not None:
            logger.trace("Wallbox: {wallbox_data.power} W", wallbox_data=wallbox_data)
            evcharger = wallbox_data.power
        elif self.settings.is_wallbox_configured:
            logger.warning("Invalid wallbox data, skipping this loop")
            logger.debug(wallbox_data)
            return

        powerflow = Powerflow(inverter_data, meters_data, batteries_data, evcharger)
        if not powerflow.is_valid:
            logger.warning("Invalid powerflow data, skipping this loop")
            logger.info(powerflow)
            return

        if Powerflow.is_not_valid_with_last(powerflow):
            logger.warning("Value change not valid, skipping this loop")
            logger.debug(powerflow)
            return

        logger.debug(powerflow)
        logger.info(
            "Powerflow: PV {pv_production} W, Inverter {inverter.power} W, "
            + "House {consumer.house} W, Grid {grid.power} W, "
            + "Battery {battery.power} W, Wallbox {consumer.evcharger} W",
            pv_production=powerflow.pv_production,
            inverter=powerflow.inverter,
            consumer=powerflow.consumer,
            grid=powerflow.grid,
            battery=powerflow.battery,
        )

        await self.mqtt.publish_components(
            inverter_data, meters_data, batteries_data, wallbox_data
        )
        await self.mqtt.publish_powerflow(powerflow)

        if self.influxdb is not None:
            self.influxdb.write_components(
                inverter_data, meters_data, batteries_data, wallbox_data
            )
            self.influxdb.write_powerflow(powerflow)

            self.influxdb.flush_loop()

    async def monitoring_loop(self):
        modules = self.monitoring.get_module_energies()

        if modules is None:
            logger.warning("Invalid monitoring data, skipping this loop")
            return

        energy_total = 0
        count_modules = 0
        for module in modules:
            if module.energy is not None:
                count_modules += 1
                energy_total += module.energy

        logger.info(
            "Read from monitoring total energy: {energy_total} kWh from {count_modules} modules",
            energy_total=energy_total / 1000,
            count_modules=count_modules,
        )

        await self.mqtt.publish_pv_energy_today(energy_total)
        await self.mqtt.publish_module_energy(modules)

    async def energy_loop(self):
        for period in EnergyPeriod:
            energy = self.influxdb.query_energy(period)
            if energy is None:
                if period.query == EnergyQuery.LAST:
                    logger.info(
                        "No data found for {period}, skipping this loop", period=period
                    )
                else:
                    logger.warning(
                        "No data found for {period}, skipping this loop", period=period
                    )
                continue

            logger.info(
                "Read from influxdb {period} energy: {energy.pv_production} kWh",
                period=period,
                energy=energy,
            )

            await self.mqtt.publish_energy(energy, period)



from __future__ import annotations

from typing import Any, Dict, Optional

from pydantic import BaseModel
from solaredge_modbus import BATTERY_STATUS_MAP, C_SUNSPEC_DID_MAP, INVERTER_STATUS_MAP

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models.base import Component, ComponentValueGroup


class SunSpecInfo(BaseModel):
    manufacturer: str
    model: str
    option: Optional[str] = None
    sunspec_type: str
    version: str
    serialnumber: str

    def __init__(self, data: Dict[str, str | int]) -> Dict[str, str]:
        values = {
            "manufacturer": data["c_manufacturer"],
            "model": data["c_model"],
            "version": data["c_version"],
            "serialnumber": data["c_serialnumber"],
        }

        if str(data["c_sunspec_did"]) in C_SUNSPEC_DID_MAP:
            values["sunspec_type"] = C_SUNSPEC_DID_MAP[str(data["c_sunspec_did"])]
        else:
            values["sunspec_type"] = "Unknown"

        if "c_option" in data:
            values["option"] = data["c_option"]

        super().__init__(**values)


class SunSpecComponent(Component):
    SOURCE = "modbus"

    info: SunSpecInfo

    def __init__(self, data: Dict[str, str | int], **kwargs):
        info = SunSpecInfo(data)
        super().__init__(info=info, **kwargs)

    def model_dump_influxdb(self, exclude: list[str] | None = None) -> Dict[str, Any]:
        return super().model_dump_influxdb(["info", *exclude] if exclude else ["info"])

    @property
    def influxdb_tags(self) -> Dict[str, str]:
        return {
            **super().influxdb_tags,
            "manufacturer": self.info.manufacturer,
            "model": self.info.model,
            "option": self.info.option,
            "sunspec_type": self.info.sunspec_type,
            "serialnumber": self.info.serialnumber,
        }


class SunSpecACCurrent(ComponentValueGroup):
    actual: float
    l1: float
    l2: Optional[float] = None
    l3: Optional[float] = None

    def __init__(self, data: Dict[str, str | int]):
        values = {"actual": self.scale_value(data, "current")}

        for phase in ["l1", "l2", "l3"]:
            if f"{phase}_current" in data:
                values[phase] = self.scale_value(
                    data, f"{phase}_current", "current_scale"
                )

        super().__init__(**values)


class SunSpecACVoltage(ComponentValueGroup):
    l1: Optional[float] = None
    l2: Optional[float] = None
    l3: Optional[float] = None
    l1n: Optional[float] = None
    l2n: Optional[float] = None
    l3n: Optional[float] = None

    def __init__(self, data: Dict[str, str | int]):
        values = {}
        for phase in ["l1", "l2", "l3"]:
            if f"{phase}_voltage" in data:
                values[phase] = self.scale_value(
                    data, f"{phase}_voltage", "voltage_scale"
                )

            if f"{phase}n_voltage" in data:
                values[f"{phase}n"] = self.scale_value(
                    data, f"{phase}n_voltage", "voltage_scale"
                )

        super().__init__(**values)


class SunSpecACPower(ComponentValueGroup):
    actual: float
    reactive: float
    apparent: float
    factor: float

    def __init__(self, data: Dict[str, str | int], power_key: str):
        actual = self.scale_value(data, power_key)
        reactive = self.scale_value(data, "power_reactive")
        apparent = self.scale_value(data, "power_apparent")
        factor = self.scale_value(data, "power_factor")
        super().__init__(
            actual=actual, reactive=reactive, apparent=apparent, factor=factor
        )


class SunSpecAC(ComponentValueGroup):
    current: SunSpecACCurrent
    voltage: SunSpecACVoltage
    power: SunSpecACPower
    frequency: float

    def __init__(self, data: Dict[str, str | int]):
        current = SunSpecACCurrent(data)
        voltage = SunSpecACVoltage(data)
        power = SunSpecACPower(data, "power_ac")
        frequency = self.scale_value(data, "frequency")

        super().__init__(
            current=current, voltage=voltage, power=power, frequency=frequency
        )


class SunSpecEnergy(ComponentValueGroup):
    totalexport: float
    totalimport: float

    def __init__(self, data: Dict[str, str | int]):
        super().__init__(
            totalexport=self.scale_value(
                data, "export_energy_active", "energy_active_scale"
            ),
            totalimport=self.scale_value(
                data, "import_energy_active", "energy_active_scale"
            ),
        )


class SunSpecDC(ComponentValueGroup):
    current: float
    voltage: float
    power: float

    def __init__(self, data: Dict[str, str | int]):
        super().__init__(
            current=self.scale_value(data, "current_dc"),
            voltage=self.scale_value(data, "voltage_dc"),
            power=self.scale_value(data, "power_dc"),
        )


class SunSpecInverter(SunSpecComponent):
    COMPONENT = "inverter"

    ac: SunSpecAC
    dc: SunSpecDC
    energytotal: float
    status: str

    def __init__(self, data: Dict[str, str | int]):
        ac = SunSpecAC(data)
        dc = SunSpecDC(data)
        energytotal = self.scale_value(data, "energy_total")
        status = INVERTER_STATUS_MAP[data["status"]]

        super().__init__(data, ac=ac, dc=dc, energytotal=energytotal, status=status)


class SunSpecMeter(SunSpecComponent):
    COMPONENT = "meter"

    current: SunSpecACCurrent
    voltage: SunSpecACVoltage
    power: SunSpecACPower
    energy: SunSpecEnergy
    frequency: float

    def __init__(self, data: Dict[str, str | int]):
        current = SunSpecACCurrent(data)
        voltage = SunSpecACVoltage(data)
        power = SunSpecACPower(data, "power")
        energy = SunSpecEnergy(data)
        frequency = self.scale_value(data, "frequency")

        super().__init__(
            data,
            current=current,
            voltage=voltage,
            power=power,
            energy=energy,
            frequency=frequency,
        )


class SunSpecBattery(SunSpecComponent):
    COMPONENT = "battery"

    status: str
    current: float
    voltage: float
    power: float
    state_of_charge: float
    state_of_health: float

    def __init__(self, data: Dict[str, str | int]):
        status = BATTERY_STATUS_MAP[data["status"]]
        current = round(data["instantaneous_current"], 2)
        voltage = round(data["instantaneous_voltage"], 2)
        power = round(data["instantaneous_power"], 2)
        state_of_charge = round(data["soe"], 2)
        state_of_health = round(data["soh"], 2)

        super().__init__(
            data,
            status=status,
            current=current,
            voltage=voltage,
            power=power,
            state_of_charge=state_of_charge,
            state_of_health=state_of_health,
        )

    @property
    def is_valid(self) -> bool:
        valid = False

        if self.state_of_charge < 0:
            logger.warning("Battery state of charge is negative")
        elif self.state_of_health < 0:
            logger.warning("Battery state of health is negative")
        elif self.current < -1000000:
            logger.warning("Battery current is a huge negative")
        else:
            valid = True

        return valid



from aiomqtt import Client, Will
from pydantic import BaseModel

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import (Component, Energy, EnergyPeriod,
                                   LogicalModule, Powerflow)
from solaredge2mqtt.settings import ServiceSettings


class MQTTClient(Client):
    def __init__(self, settings: ServiceSettings):
        self.broker = settings.broker
        self.port = settings.port

        self.topic_prefix = settings.topic_prefix

        logger.info(
            "Using MQTT broker: {broker}:{port}",
            broker=settings.broker,
            port=settings.port,
        )

        will = Will(
            topic=f"{self.topic_prefix}/status", payload="offline", qos=1, retain=True
        )

        super().__init__(
            self.broker,
            self.port,
            username=settings.username,
            password=settings.password,
            client_id=settings.client_id,
            will=will,
        )

    async def publish_status_online(self) -> None:
        await self.publish(f"{self.topic_prefix}/status", "online", qos=1, retain=True)

    async def publish_status_offline(self) -> None:
        await self.publish(f"{self.topic_prefix}/status", "offline", qos=1, retain=True)

    async def publish_components(
        self, *args: list[Component | dict[str, Component] | None]
    ) -> None:
        for arg in args:
            if isinstance(arg, dict):
                for component_key, component in arg.items():
                    await self._publish(
                        f"{component.SOURCE}/{component.COMPONENT}/{component_key.lower()}",
                        component,
                    )
            elif isinstance(arg, Component):
                await self._publish(f"{arg.SOURCE}/{arg.COMPONENT}", arg)
            elif arg is None:
                continue
            else:
                raise ValueError(f"Invalid component: {arg}")

    async def publish_powerflow(self, powerflow: Powerflow) -> None:
        await self._publish("powerflow", powerflow)

    async def publish_energy(self, energy: Energy, period: EnergyPeriod) -> None:
        await self._publish(f"energy/{period.topic}", energy)

    async def publish_pv_energy_today(self, energy: int) -> None:
        await self._publish("api/monitoring/pv_energy_today", energy)

    async def publish_module_energy(self, modules: list[LogicalModule]) -> None:
        for module in modules:
            await self._publish(
                f"api/monitoring/module/{module.info.serialnumber}",
                module,
            )

    async def _publish(
        self, topic: str, payload: str | int | float | BaseModel
    ) -> None:
        if self._connected:
            if isinstance(payload, BaseModel):
                payload = payload.model_dump_json()
            await self.publish(f"{self.topic_prefix}/{topic}", payload, qos=1)



import json
from typing import Dict, Tuple

from pymodbus.exceptions import ModbusException

from solaredge_modbus import Inverter

from solaredge2mqtt.logging import LOGGING_DEVICE_INFO, logger
from solaredge2mqtt.models import (
    Powerflow,
    SunSpecBattery,
    SunSpecInverter,
    SunSpecMeter,
)
from solaredge2mqtt.settings import ServiceSettings

SunSpecRawData = Dict[str, str | int]


class Modbus:
    inverter: Inverter

    def __init__(self, settings: ServiceSettings):
        self.inverter = Inverter(
            host=settings.modbus_host,
            port=settings.modbus_port,
            timeout=settings.modbus_timeout,
            unit=settings.modbus_unit,
        )

        logger.info(
            "Using SolarEdge inverter via modbus: {host}:{port}",
            host=settings.modbus_host,
            port=settings.modbus_port,
        )

    async def loop(
        self,
    ) -> Tuple[
        SunSpecInverter | None,
        Dict[str, SunSpecMeter] | None,
        Dict[str, SunSpecBattery | None],
    ] | None:
        inverter_data = None
        meters_data = None
        batteries_data = None

        try:
            inverter_raw, meters_raw, batteries_raw = self._get_raw_data()

            inverter_data = self._map_inverter(inverter_raw)
            meters_data = self._map_meters(meters_raw)
            batteries_data = self._map_batteries(batteries_raw)
        except ModbusException as error:
            logger.error(
                "Exception while reading data from modbus: {exception}",
                exception=error,
            )

        return inverter_data, meters_data, batteries_data

    def _get_raw_data(
        self,
    ) -> Tuple[SunSpecRawData, Dict[str, SunSpecRawData], Dict[str, SunSpecRawData]]:
        inverter_raw = self.inverter.read_all()
        meters_raw = {
            meter_key: meter_obj.read_all()
            for meter_key, meter_obj in self.inverter.meters().items()
        }
        batteries_raw = {
            battery_key: battery_obj.read_all()
            for battery_key, battery_obj in self.inverter.batteries().items()
        }

        return inverter_raw, meters_raw, batteries_raw

    def _map_inverter(self, inverter_raw: SunSpecRawData) -> SunSpecInverter:
        logger.debug(
            "Inverter raw:\n{raw}",
            raw=json.dumps(inverter_raw, indent=4),
        )

        inverter_data = SunSpecInverter(inverter_raw)
        logger.debug(inverter_data)
        logger.info(
            LOGGING_DEVICE_INFO
            + ": {status}, AC {power_ac} W, DC {power_dc} W, {energytotal} kWh",
            device="Inverter",
            info=inverter_data.info,
            status=inverter_data.status,
            power_ac=inverter_data.ac.power.actual,
            power_dc=inverter_data.dc.power,
            energytotal=round(inverter_data.energytotal / 1000, 2),
        )

        return inverter_data

    def _map_meters(
        self, meters_raw: Dict[str, SunSpecRawData]
    ) -> Dict[str, SunSpecInverter]:
        meters = {}
        for meter_key, meter_raw in meters_raw.items():
            logger.debug(
                "Meter {meter} raw:\n{raw}",
                meter=meter_key,
                raw=json.dumps(meter_raw, indent=4),
            )

            meter_data = SunSpecMeter(meter_raw)
            logger.debug(meter_data)
            logger.info(
                LOGGING_DEVICE_INFO + ": {power} W, {consumption} kWh, {delivery} kWh",
                device=meter_key,
                info=meter_data.info,
                power=meter_data.power.actual,
                consumption=round(meter_data.energy.totalimport / 1000, 3),
                delivery=round(meter_data.energy.totalexport / 1000, 3),
            )

            meters[meter_key] = meter_data

        return meters

    def _map_batteries(
        self, batteries_raw: Dict[str, SunSpecRawData]
    ) -> Dict[str, SunSpecInverter]:
        batteries = {}
        for battery_key, battery_raw in batteries_raw.items():
            logger.debug(
                "Battery {battery} raw:\n{raw}",
                battery=battery_key,
                raw=json.dumps(battery_raw, indent=4),
            )

            battery_data = SunSpecBattery(battery_raw)
            logger.debug(battery_data)
            logger.info(
                LOGGING_DEVICE_INFO + ": {power} W, {state_of_charge} %",
                device=battery_key,
                info=battery_data.info,
                power=battery_data.power,
                state_of_charge=battery_data.state_of_charge,
            )

            batteries[battery_key] = battery_data

        return batteries

    @staticmethod
    def calc_powerflow(
        inverter: SunSpecInverter,
        meters: Dict[str, SunSpecMeter],
        batteries: Dict[str, SunSpecBattery],
    ) -> Powerflow:
        """
        Calculates the power flow in the system by summing the power of all meters and batteries.
        It considers both import and export options for each meter in the calculation.
        Returns a PowerFlow object representing the total power flow in the system.
        """

        powerflow = Powerflow.calc(inverter, meters, batteries)

        logger.debug(powerflow)
        logger.info(
            "Powerflow: PV {pv_production} W, Inverter {inverter.power} W, "
            + "House {house_consumption} W, "
            + "Grid {grid.power} W, Battery {battery.power} W",
            pv_production=powerflow.pv_production,
            inverter=powerflow.inverter,
            house_consumption=powerflow.house_consumption,
            grid=powerflow.grid,
            battery=powerflow.battery,
        )

        return powerflow



from __future__ import annotations

from typing import ClassVar, Dict, Optional

from pydantic import Field, computed_field

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models.base import Solaredge2MQTTBaseModel
from solaredge2mqtt.models.modbus import SunSpecBattery, SunSpecInverter, SunSpecMeter


class InverterPowerflow(Solaredge2MQTTBaseModel):
    power: int
    dc_power: int
    battery_factor: float = Field(exclude=True)

    def __init__(
        self,
        inverter_data: SunSpecInverter,
        battery: BatteryPowerflow,
    ):
        power = int(inverter_data.ac.power.actual)
        dc_power = int(inverter_data.dc.power)
        battery_factor = 0.0

        if power > 0 and battery.discharge > 0:
            battery_factor = battery.discharge / dc_power

        super().__init__(
            power=power,
            dc_power=dc_power,
            battery_factor=battery_factor,
        )

    @computed_field
    @property
    def consumption(self) -> int:
        return abs(self.power) if self.power < 0 else 0

    @computed_field
    @property
    def production(self) -> int:
        return self.power if self.power > 0 else 0

    @computed_field
    @property
    def battery_production(self) -> int:
        battery_production = 0
        if self.production > 0 and self.battery_factor > 0:
            battery_production = int(round(self.production * self.battery_factor))
            battery_production = min(battery_production, self.production)
        return battery_production

    @computed_field
    @property
    def pv_production(self) -> int:
        return self.production - self.battery_production

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.consumption < 0:
            logger.warning("Inverter consumption is negative")
        elif self.production < 0:
            logger.warning("Inverter production is negative")
        elif self.pv_production < 0:
            logger.warning("Inverter PV production is negative")
        elif self.battery_production < 0:
            logger.warning("Inverter battery production is negative")
        else:
            valid = True

        return valid


class GridPowerflow(Solaredge2MQTTBaseModel):
    power: int

    def __init__(self, meters_data: Dict[str, SunSpecMeter]):
        grid = 0
        for meter in meters_data.values():
            if "Import" in meter.info.option and "Export" in meter.info.option:
                grid += meter.power.actual

        super().__init__(power=grid)

    @computed_field
    @property
    def consumption(self) -> int:
        return abs(self.power) if self.power < 0 else 0

    @computed_field
    @property
    def delivery(self) -> int:
        return self.power if self.power > 0 else 0

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.consumption < 0:
            logger.warning("Grid consumption is negative")
        elif self.delivery < 0:
            logger.warning("Grid delivery is negative")
        else:
            valid = True

        return valid


class BatteryPowerflow(Solaredge2MQTTBaseModel):
    power: int

    def __init__(self, batteries_data: Dict[str, SunSpecBattery]):
        batteries_power = 0
        for battery in batteries_data.values():
            batteries_power += battery.power

        super().__init__(power=batteries_power)

    @computed_field
    @property
    def charge(self) -> int:
        return self.power if self.power > 0 else 0

    @computed_field
    @property
    def discharge(self) -> int:
        return abs(self.power) if self.power < 0 else 0

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.charge < 0:
            logger.warning("Battery charge is negative")
        elif self.discharge < 0:
            logger.warning("Battery discharge is negative")
        else:
            valid = True

        return valid


class ConsumerPowerflow(Solaredge2MQTTBaseModel):
    house: int
    evcharger: int = 0
    inverter: int

    used_production: int
    battery_factor: float = Field(exclude=True)

    def __init__(
        self, inverter: InverterPowerflow, grid: GridPowerflow, evcharger: int
    ):
        house = int(abs(grid.power - inverter.power)) - evcharger

        battery_factor = inverter.battery_factor

        if inverter.production > 0 and inverter.production > grid.delivery:
            used_production = inverter.production - grid.delivery
        else:
            used_production = 0

        super().__init__(
            house=house,
            evcharger=evcharger,
            inverter=inverter.consumption,
            used_production=used_production,
            battery_factor=battery_factor,
        )

    @computed_field
    @property
    def total(self) -> int:
        return self.house + self.evcharger + self.inverter

    @computed_field
    @property
    def used_battery_production(self) -> int:
        battery_production = 0
        if self.used_production > 0 and self.battery_factor > 0:
            battery_production = int(round(self.used_production * self.battery_factor))
            battery_production = min(battery_production, self.used_production)
        return battery_production

    @computed_field
    @property
    def used_pv_production(self) -> int:
        return self.used_production - self.used_battery_production

    @property
    def is_valid(self) -> bool:
        valid = False
        if self.house < 0:
            logger.warning("Consumer house is negative")
        elif self.evcharger < 0:
            logger.warning("Consumer evcharger is negative")
        elif self.inverter < 0:
            logger.warning("Consumer inverter is negative")
        elif self.used_production < 0:
            logger.warning("Consumer used production is negative")
        elif self.used_pv_production < 0:
            logger.warning("Consumer used PV production is negative")
        elif self.used_battery_production < 0:
            logger.warning("Consumer used battery production is negative")
        elif self.total < 0:
            logger.warning("Consumer total is negative")
        elif self.total < self.used_production:
            logger.warning("Consumer total is less than used production")
        else:
            valid = True

        return valid


class Powerflow(Solaredge2MQTTBaseModel):
    pv_production: int
    inverter: InverterPowerflow
    grid: GridPowerflow
    battery: BatteryPowerflow
    consumer: ConsumerPowerflow

    last_powerflow: ClassVar[Powerflow] = None

    def __init__(
        self,
        inverter_data: SunSpecInverter,
        meters_data: Dict[str, SunSpecMeter],
        batteries_data: Dict[str, SunSpecBattery],
        evcharger: Optional[int] = 0,
    ):
        grid = GridPowerflow(meters_data)
        battery = BatteryPowerflow(batteries_data)

        if inverter_data.ac.power.actual > 0:
            pv_production = int(inverter_data.dc.power + battery.power)
            if pv_production < 0:
                pv_production = 0
        else:
            pv_production = 0

        inverter = InverterPowerflow(inverter_data, battery)

        consumer = ConsumerPowerflow(inverter, grid, evcharger)

        super().__init__(
            pv_production=pv_production,
            inverter=inverter,
            grid=grid,
            battery=battery,
            consumer=consumer,
        )

    @property
    def is_valid(self) -> bool:
        valid = False

        if self.pv_production < 0:
            logger.warning("PV production is negative")
        elif (
            self.consumer.used_production + self.grid.delivery
            != self.inverter.production
        ):
            logger.warning(
                "Consumer used production + grid delivery is not equal to inverter production"
            )
        else:
            valid = all(
                [
                    self.inverter.is_valid,
                    self.grid.is_valid,
                    self.battery.is_valid,
                    self.consumer.is_valid,
                ]
            )

        return valid

    @classmethod
    def is_not_valid_with_last(cls, powerflow: Powerflow) -> bool:
        check = False

        if cls.last_powerflow is not None:
            check = all(
                [
                    cls.last_powerflow.pv_production == 0
                    and powerflow.pv_production > 100,
                ]
            )

        cls.last_powerflow = powerflow

        return check



import pkg_resources
from influxdb_client import (BucketRetentionRules, BucketsApi, InfluxDBClient,
                             Point, TaskCreateRequest)
from influxdb_client.client.exceptions import InfluxDBError

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import Component, Energy, EnergyPeriod, Powerflow
from solaredge2mqtt.settings import ServiceSettings


class InfluxDB:
    def __init__(self, settings: ServiceSettings) -> None:
        self.settings = settings

        self.org: str = settings.influxdb_org

        self.client: InfluxDBClient = InfluxDBClient(
            url=f"{settings.influxdb_host}:{settings.influxdb_port}",
            token=settings.influxdb_token,
            org=settings.influxdb_org,
        )

        self.write_api = self.client.write_api(
            success_callback=self.write_success_callback,
            error_callback=self.write_error_callback,
            retry_callback=self.write_error_callback,
        )

        self.query_api = self.client.query_api()

        self.loop_points: list[Point] = []

        self.flux_cache: dict[str, str] = {}

    def initialize_buckets(self) -> None:
        buckets_api = self.client.buckets_api()
        self._create_or_update_bucket(
            buckets_api, self.bucket_raw, self.settings.influxdb_retention_raw
        )
        self._create_or_update_bucket(
            buckets_api,
            self.bucket_aggregated,
            self.settings.influxdb_retention_aggregated,
        )

    def _create_or_update_bucket(
        self, buckets_api: BucketsApi, bucket_name: str, retention: int
    ) -> None:
        bucket = buckets_api.find_bucket_by_name(bucket_name)
        retention_rules = BucketRetentionRules(type="expire", every_seconds=retention)

        if bucket is None:
            logger.info(f"Creating bucket '{bucket_name}'")
            buckets_api.create_bucket(
                bucket_name=bucket_name, retention_rules=retention_rules
            )
        else:
            logger.info(f"Bucket '{bucket_name}' already exists.")
            if bucket.retention_rules[0].every_seconds != retention:
                logger.info(
                    f"Updating retention rules for bucket '{bucket_name}' to {retention} seconds."
                )
                bucket.retention_rules[0] = retention_rules
                buckets_api.update_bucket(bucket=bucket)

    @property
    def bucket_raw(self) -> str:
        return f"{self.settings.influxdb_prefix}_raw"

    @property
    def bucket_aggregated(self) -> str:
        return f"{self.settings.influxdb_prefix}"

    def initialize_task(self) -> None:
        tasks_api = self.client.tasks_api()
        tasks = tasks_api.find_tasks(name=self.task_name)
        flux = (
            self._get_flux_query("aggregation")
            .replace("TASK_NAME", self.task_name)
            .replace("UNIT", self.settings.influxdb_aggregate_interval)
        )

        if not tasks:
            task_request = TaskCreateRequest(
                flux=flux, org_id=self.org, status="active"
            )
            logger.info(f"Creating task '{self.task_name}'")
            logger.debug(flux)
            tasks_api.create_task(task_create_request=task_request)
        else:
            logger.info(f"Task '{self.task_name}' already exists.")

            new_flux = self._strip_flux(flux)
            stored_flux = self._strip_flux(tasks[0].flux)

            if (
                new_flux != stored_flux
                or tasks[0].every != self.settings.influxdb_aggregate_interval
            ):
                logger.info(f"Updating task '{self.task_name}'")
                logger.debug(flux)
                tasks[0].flux = flux
                tasks[0].every = self.settings.influxdb_aggregate_interval
                tasks_api.update_task(tasks[0])

    @staticmethod
    def _strip_flux(flux: str) -> str:
        return "".join(line.strip() for line in flux.splitlines()).replace(" ", "")

    def _get_flux_query(self, query_name: str) -> str:
        if query_name not in self.flux_cache:
            flux = pkg_resources.resource_string(
                __name__, f"resources/{query_name}.flux"
            ).decode("utf-8")
            flux = (
                flux.replace("BUCKET_RAW", self.bucket_raw)
                .replace("BUCKET_AGGREGATED", self.bucket_aggregated)
                .replace("TIMEZONE", self.settings.timezone)
            )
            self.flux_cache[query_name] = flux

        return self.flux_cache[query_name]

    @property
    def task_name(self) -> str:
        return f"{self.settings.influxdb_prefix}_aggregation"

    def write_components(
        self, *args: list[Component | dict[str, Component] | None]
    ) -> None:
        for component in args:
            if isinstance(component, Component):
                self.loop_points.append(self._create_component_point(component))
            elif isinstance(component, dict):
                for name, component in component.items():
                    self.loop_points.append(
                        self._create_component_point(component, {"name": name})
                    )

    def write_component(self, component: Component) -> None:
        self.loop_points.append(self._create_component_point(component))

    def _create_component_point(
        self, component: Component, additional_tags: dict[str, str] = None
    ) -> Point:
        point = Point("component")

        for key, value in component.influxdb_tags.items():
            point.tag(key, value)

        if additional_tags is not None:
            for key, value in additional_tags.items():
                point.tag(key, value)

        for key, value in component.model_dump_influxdb().items():
            point.field(key, value)

        return point

    def write_powerflow(self, powerflow: Powerflow) -> None:
        point = Point("powerflow")
        for key, value in powerflow.model_dump_influxdb().items():
            point.field(key, value)
        self.loop_points.append(point)

    def flush_loop(self) -> None:
        self.write_api.write(bucket=self.bucket_raw, record=self.loop_points)
        self.loop_points = []

    def write_success_callback(self, conf: (str, str, str), data: str) -> None:
        logger.debug(f"InfluxDB batch written: {conf} {data}")

    def write_error_callback(
        self, conf: (str, str, str), data: str, error: InfluxDBError
    ) -> None:
        logger.error(f"InfluxDB error while writting: {conf} {error}")
        logger.debug(data)

    def query_energy(self, period: EnergyPeriod) -> Energy | None:
        energy: Energy = None
        query = self._get_flux_query(period.query.query).replace("UNIT", period.unit)
        logger.trace(query)
        tables = self.query_api.query(query)
        for table in tables:
            for record in table.records:
                logger.trace(record)
                energy = Energy(record.values, period)

        return energy



from __future__ import annotations
from collections.abc import MutableMapping
from typing import Any, Dict, Optional, ClassVar

from enum import Enum

from pydantic import BaseModel, model_serializer


class EnumModel(Enum):
    def __new__(cls, *args):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __str__(self) -> str:
        return str(self._value_)

    def __repr__(self) -> str:
        return str(self._value_)

    def __eq__(self, obj) -> bool:
        return isinstance(obj, self.__class__) and self.value == obj.value

    def __hash__(self):
        return hash(self.value)

    @classmethod
    def from_string(cls, value: str):
        for item in cls.__members__.values():
            if item.value == value:
                return item
        raise ValueError(f"No enum value {value} found.")

    @model_serializer
    def serialize(self):
        return self.value


class Solaredge2MQTTBaseModel(BaseModel):
    def model_dump_influxdb(
        self, exclude: Optional[list[str]] = None
    ) -> Dict[str, Any]:
        return self._flatten_dict(self.model_dump(exclude=exclude, exclude_none=True))

    def _flatten_dict(self, d: MutableMapping, parent_key: str = "") -> MutableMapping:
        items = []
        for k, v in d.items():
            new_key = parent_key + "_" + k if parent_key else k
            if isinstance(v, MutableMapping):
                items.extend(self._flatten_dict(v, new_key).items())
            else:
                items.append((new_key, float(v) if isinstance(v, int) else v))
        return dict(items)


class ComponentValueGroup(Solaredge2MQTTBaseModel):
    @staticmethod
    def scale_value(
        data: Dict[str, str | int],
        value_key: str,
        scale_key: Optional[str] = None,
        digits: Optional[int] = 2,
    ) -> float:
        if scale_key is None:
            scale_key = f"{value_key}_scale"

        value = int(data[value_key])
        scale = int(data[scale_key])

        return round(value * 10**scale, digits)


class Component(ComponentValueGroup):
    COMPONENT: ClassVar[str] = "unknown"
    SOURCE: ClassVar[str] = "unknown"

    @property
    def influxdb_tags(self) -> Dict[str, str]:
        return {
            "component": self.COMPONENT,
            "source": self.SOURCE,
        }



from functools import lru_cache
from os import path
from time import localtime, strftime
from typing import Optional

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from solaredge2mqtt.logging import LoggingLevelEnum
from solaredge2mqtt.models import EnumModel

DOCKER_SECRETS_DIR = "/run/secrets"

MODEL_CONFIG_WITHOUT_SECRETS = {"env_file": ".env", "env_prefix": "se2mqtt_"}

MODEL_CONFIG_WITH_SECRETS = {
    **MODEL_CONFIG_WITHOUT_SECRETS,
    "secrets_dir": DOCKER_SECRETS_DIR,
}

MODEL_CONFIG = (
    SettingsConfigDict(**MODEL_CONFIG_WITH_SECRETS)
    if path.exists(DOCKER_SECRETS_DIR)
    else SettingsConfigDict(**MODEL_CONFIG_WITHOUT_SECRETS)
)


SECONDS_PER_DAY = 86400
SECONDS_PER_HOUR = 3600
SECONDS_PER_YEAR = SECONDS_PER_DAY * 365
SECONDS_PER_2_YEARS = SECONDS_PER_YEAR * 2


class ServiceSettings(BaseSettings):
    environment: str = "production"

    modbus_host: str
    modbus_port: int = Field(1502)
    modbus_timeout: int = Field(1)
    modbus_unit: int = Field(1)

    api_site_id: Optional[str] = Field(None)
    api_username: Optional[str] = Field(None)
    api_password: Optional[str] = Field(None)

    client_id: str = Field("solaredge2mqtt")
    broker: str
    port: int = Field(1883)
    username: str
    password: str
    topic_prefix: str = Field("solaredge")

    interval: int = Field(5)

    wallbox_host: Optional[str] = Field(None)
    wallbox_password: Optional[str] = Field(None)
    wallbox_serial: Optional[str] = Field(None)

    influxdb_host: Optional[str] = Field(None)
    influxdb_port: Optional[int] = Field(8086)
    influxdb_token: Optional[str] = Field(None)
    influxdb_org: Optional[str] = Field(None)
    influxdb_prefix: Optional[str] = Field("solaredge")
    influxdb_retention_raw: Optional[int] = Field(SECONDS_PER_DAY + SECONDS_PER_HOUR)
    influxdb_retention_aggregated: Optional[int] = Field(SECONDS_PER_2_YEARS)
    influxdb_aggregate_interval: Optional[str] = "10m"

    logging_level: LoggingLevelEnum = LoggingLevelEnum.INFO
    timezone: str = Field(strftime("%Z", localtime()))

    model_config = MODEL_CONFIG

    @property
    def is_api_configured(self) -> bool:
        return all(
            [
                self.api_site_id is not None,
                self.api_username is not None,
                self.api_password is not None,
            ]
        )

    @property
    def is_wallbox_configured(self) -> bool:
        return all(
            [
                self.wallbox_host is not None,
                self.wallbox_password is not None,
                self.wallbox_serial is not None,
            ]
        )

    @property
    def is_influxdb_configured(self) -> bool:
        return all(
            [
                self.influxdb_host is not None,
                self.influxdb_port is not None,
                self.influxdb_token is not None,
                self.influxdb_org is not None,
            ]
        )


class DevelopmentSettings(ServiceSettings):
    debug: bool = Field(True)

    environment: str = "development"

    influxdb_prefix: Optional[str] = Field("solaredgedev")

    model_config = MODEL_CONFIG


class ServiceEnvironment(EnumModel):
    PROD = "production", ServiceSettings
    DEV = "development", DevelopmentSettings

    def __init__(self, description: str, settings_class: ServiceSettings):
        # pylint: disable=super-init-not-called
        self._description: str = description
        self._settings_class: ServiceSettings = settings_class

    @property
    def description(self) -> str:
        return self._description

    @property
    def settings_class(self) -> ServiceSettings:
        return self._settings_class


class ServiceBaseSettings(BaseSettings):
    environment: ServiceEnvironment = ServiceEnvironment.PROD

    model_config = SettingsConfigDict(**MODEL_CONFIG_WITHOUT_SECRETS, extra="ignore")


@lru_cache()
def service_settings() -> ServiceSettings:
    environment = ServiceBaseSettings().environment
    return environment.settings_class()



from os import path

import setuptools

import versioneer

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.readlines()


packages = setuptools.find_packages(exclude=["tests"])

setuptools.setup(
    name="solaredge2mqtt",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Read data from SolarEdge Inverter and publish it to MQTT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deroetzi/solaredge2mqtt",
    author="Johannes Ott",
    author_email="info@johannes-ott.net",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Home Automation",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="smart home",
    packages=packages,
    include_package_data=True,
    python_requires=">=3.10, <4",
    install_requires=[req for req in requirements if not req.startswith("#")],
    entry_points={"console_scripts": ["solaredge2mqtt=solaredge2mqtt.service:run"]},
    project_urls={"Bug Reports": "https://github.com/deroetzi/solaredge2mqtt/issues"},
)



import time
from typing import Optional

import jwt
import urllib3
from pydantic import BaseModel, Field
from requests import get, post
from requests.exceptions import ConnectionError as RequestsConnectionError
from requests.exceptions import HTTPError, Timeout

from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import WallboxAPI
from solaredge2mqtt.settings import ServiceSettings

urllib3.disable_warnings()

LOGIN_URL = "https://{host}:8443/v2/jwt/login"
REFRESH_URL = "https://{host}:8443/v2/jwt/refresh"
WALLBOX_URL = "https://{host}:8443/v2/wallboxes/{serial}"


class AuthorizationTokens(BaseModel):
    access_token: str = Field(None, alias="accessToken")
    refresh_token: str = Field(None, alias="refreshToken")

    @property
    def access_token_expires(self) -> int:
        payload = jwt.decode(self.access_token, options={"verify_signature": False})
        logger.debug("Access token expires at: {expires}", expires=payload["exp"])
        return payload["exp"]

    @property
    def refresh_token_expires(self) -> int:
        payload = jwt.decode(self.refresh_token, options={"verify_signature": False})
        logger.debug("Refresh token expires at: {expires}", expires=payload["exp"])
        return payload["exp"]


class WallboxClient:
    def __init__(self, settings: ServiceSettings):
        self.host = settings.wallbox_host
        self.password = settings.wallbox_password
        self.serial = settings.wallbox_serial

        logger.info(
            "Using Wallbox charger: {host}",
            host=settings.wallbox_host,
        )

        self.authorization: Optional[AuthorizationTokens] = None

    async def loop(self) -> WallboxAPI | None:
        wallbox = None

        try:
            self._get_access()

            response = get(
                WALLBOX_URL.format(host=self.host, serial=self.serial),
                headers={"Authorization": f"Bearer {self.authorization.access_token}"},
                timeout=5,
                verify=False,  # pylint: disable=insecure-request-warning
            )

            if response.status_code == 401:
                # Access token is invalid, login again (may happen after restart)
                self._login()
                wallbox = self.loop()
            else:
                response.raise_for_status()
                wallbox = WallboxAPI(response.json())
                logger.info(
                    f"Wallbox: {wallbox.state}, {wallbox.power / 1000} kW",
                    wallbox=wallbox,
                )
        except HTTPError as error:
            logger.error(
                "Failed to get Wallbox data: {error}",
                error=error,
            )
        except RequestsConnectionError as error:
            logger.error(
                "Failed to connect to Wallbox charger: {error}",
                error=error,
            )
        except Timeout as error:
            logger.error(
                "Timeout while connecting to Wallbox charger: {error}",
                error=error,
            )

        return wallbox

    def _get_access(self) -> None:
        current_timestamp = int(time.time())

        try:
            if self.authorization is None:
                self._login()
            elif self.authorization.access_token_expires < current_timestamp + 60:
                # Token is about to expire within 60 seconds
                if self.authorization.refresh_token_expires < current_timestamp + 60:
                    # Refresh token is about to expire within 60 seconds as well new login
                    self._login()
                else:
                    self._refresh_token()
            else:
                logger.debug("Wallbox access token still valid")

        except HTTPError as error:
            logger.error(
                "Failed to get access token for wallbox charger: {error}",
                error=error,
            )
            raise error

    def _login(self):
        logger.info("Logging in to Wallbox charger...")
        self.authorization = None
        response = post(
            LOGIN_URL.format(host=self.host),
            json={"password": self.password, "username": "admin"},
            timeout=5,
            verify=False,  # pylint: disable=insecure-request-warning
        )
        response.raise_for_status()
        logger.trace(response.json())
        self.authorization = AuthorizationTokens(**response.json())
        logger.info("Logged in to EV charger")

    def _refresh_token(self):
        logger.info("Refreshing access token Wallbox...")
        response = post(
            REFRESH_URL.format(host=self.host),
            headers={"Authorization": f"Bearer {self.authorization.refresh_token}"},
            timeout=5,
            verify=False,  # pylint: disable=insecure-request-warning
        )

        if response.status_code == 401:
            # Refresh token is invalid, login again
            self._login()
        else:
            response.raise_for_status()
            payload = response.json()
            self.authorization.access_token = payload["accessToken"]

        logger.info("Refreshed access token Wallbox")
